{% extends "!layout.html" %}

 {%- block extrahead %}
 {{ super() }}
    <script type="text/javascript" src="_static/copybutton.js"></script>
 {% endblock %}

 {%- block footer %}
    {{super()}}
    <span id="forkongithub"><a href="https://github.com/{{ theme_github_user }}/{{ theme_github_repo }}">Fork me on GitHub</a></span>
 {%- endblock %}
